#
# Copyright (c) 2015-2016,2019 LAAS/CNRS
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software for any purpose
# with or without   fee is hereby granted, provided   that the above  copyright
# notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS  SOFTWARE INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR  BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR  ANY DAMAGES WHATSOEVER RESULTING  FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION,   ARISING OUT OF OR IN    CONNECTION WITH THE USE   OR
# PERFORMANCE OF THIS SOFTWARE.
#
#                                             Anthony Mallet on Mon May 18 2015
#
import errno
import json
import re
import threading
import urlparse
import BaseHTTPServer
import SocketServer

import node
import session
from util import *
from version import __genomix_protocol__

_listener = None
_httpd = None

_push_channels = {}


# --- start ----------------------------------------------------------------
#
def start(port):
  global _listener, _httpd

  _httpd = _http_server(('', port), _http_handler)
  _listener = threading.Thread(target=_httpd.serve_forever)
  _listener.start()


# --- stop -----------------------------------------------------------------
#
def stop():
  global _listener, _httpd

  if _httpd is not None:
    _httpd.shutdown()
  if _listener is not None:
    _listener.join()

  _listener = None
  _httpd = None


# --- push -----------------------------------------------------------------
#
def push(channel, event, data = ''):
  global _push_channels

  if channel not in _push_channels: return
  r = _push_channels[channel]['handler']

  infox('%s: pushing event: %s data: %s' % (channel, event, data), 1)

  e = 'event: %s\n' % event
  for l in data.splitlines():
    e = e + 'data: %s\n' % l
  e = e + '\n'

  try:
    r.wfile.write(e)
    r.wfile.flush()
  except:
    _push_channels[channel]['event'].set()
  else:
    if event == 'quit':
      _push_channels[channel]['event'].set()


# --- http_handler ---------------------------------------------------------
#
class _http_server(SocketServer.ThreadingMixIn,
                   BaseHTTPServer.HTTPServer):
  daemon_threads = True
  pass

class _http_handler(BaseHTTPServer.BaseHTTPRequestHandler):
  channel = 0
  protocol_version = 'HTTP/1.1'

  def __init__(self, r, addr, s):
    _http_handler.channel += 1
    self.channel = 'sock%d' % _http_handler.channel
    infox('connection from %s:%d on %s' % (addr[0], addr[1], self.channel))
    try:
      BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, r, addr, s)
    except IOError as e:
      if e.errno == errno.EPIPE:
        pass # ignore
      else:
        raise

  def reply(self, status, msg):
    infox('%s: reply status %d: %s' % (self.channel, status, msg), 1)

    self.send_response(status)
    self.send_header('Connection', 'keep-alive')
    self.send_header('Content-Type', 'application/json')
    self.send_header('Content-Length', len(msg))
    self.end_headers()
    self.wfile.write(msg)


  def stream(self):
    global _push_channels

    done = threading.Event()
    _push_channels[self.channel] = { 'handler': self, 'event': done }

    infox("%s: set to streaming mode" % self.channel, 1)

    self.send_response(200)
    self.send_header('Connection', 'keep-alive')
    self.send_header('Content-Type', 'text/event-stream')
    self.end_headers()
    self.wfile.flush()

    done.wait()
    del _push_channels[self.channel]


  def do_get(self):
    infox('%s: "%s"' % (self.channel, self.requestline), 1)

    url = urlparse.urlparse(self.path)
    args = urlparse.parse_qs(url.query, True)

    if re.match(r'^/session/', url.path):
      try:
        c = url.path.split('/')[2]
        r = session.keys[c](self.channel, args)
        if r is not None:
          self.reply(200, r)
        else:
          self.stream()

      except softerror as e:
        self.reply(400, '%s' % e.args)
      except KeyError as e:
        self.reply(404, json.dumps('%s' % url.path))

    elif re.match(r'^/load/', url.path):
      try:
        p = '/'.join(url.path.split('/')[2:])
        if 'argv' not in args: args['argv'] = []
        r = node.load(p, args['argv'])
        self.reply(201, r)

      except softerror as e:
        self.reply(404, '%s' % e.args)

    elif re.match(r'^/module/', url.path):
      u = url.path.split('/')
      m = node.get_proxy(u[2])
      c = u[3]

      if m is None:
        self.reply(404, u[2])
        return

      try:
        if c == 'unload':
          session.broadcast('unload', m.name)
          del(m)
          r = json.dumps('unloaded %s' % u[2])

        elif c == 'send':
          if 'input' in args:
            inp = json.loads(args['input'][0])
          else:
            inp = {}

          r = m.send('/'.join(u[4:]), inp, args)

        elif c == 'read':
          r = m.read('/'.join(u[4:]))

        elif c in ('status', 'abort', 'clean'):
          r = getattr(m, c)(int(u[4]))

        else:
          if c not in m.keys:
            self.reply(400, json.dumps('Bad request: "%s"' % c))
            return

          r = getattr(m, c)()

        self.reply(200, r)
      except softerror as e:
        self.reply(404, '%s' % e.args)

    elif re.match(r'^/usage/', url.path):
      self.reply(200, '[]')

    elif re.match(r'^/rpath/', url.path):
      self.reply(200, '')

    elif re.match(r'^/version$', url.path):
      self.reply(200, __genomix_protocol__)

    else:
      self.reply(400, json.dumps('Bad request: "%s"' % url.path))

  do_GET = do_get

  def log_request(self, code='-', size='-'):
    pass

  def log_message(self, format, *args):
    infox(format % args)
